"use strict";
exports.__esModule = true;
exports.StorageContext = void 0;
var CacheStrategy_1 = require("../base/CacheStrategy");
var iterables_1 = require("../iterables");
var LeafStorage_1 = require("./LeafStorage");
var storage_1 = require("../storage");
var RootLeafStrategy_1 = require("./RootLeafStrategy");
var strategy_1 = require("../strategy");
var constants_1 = require("../constants");
var utils_1 = require("../utils");
/**
 * Picks a `Noop` strategy if the `limit` is `Infinity`.
 * @param strategy
 * @param limit
 */
var pickNoopIfNoLimit = function (strategy, limit) { return (limit !== Infinity ? strategy : strategy_1.Noop); };
/**
 * Checks whether the provided limit is a valid number.
 * @param limit
 */
var limitIsValid = function (limit) {
    return (Number.isSafeInteger(limit) && limit > 0) || limit === Infinity;
};
/**
 * Ensures that the provided limit is a valid number.
 * @param name
 * @param value
 * @param min
 */
var ensureLimitIsValid = function (name, value, min) {
    if (!limitIsValid(value)) {
        throw new Error("`".concat(name, "` is invalid: ").concat(value));
    }
    if (value < min) {
        throw new Error("".concat(value, " is too small, expected ").concat(min, " at least"));
    }
};
/**
 * Returns `true` if supplied class extends `CacheStrategy`.
 * @param strategy
 */
var extendsCacheStrategy = function (strategy) {
    return strategy &&
        strategy.prototype instanceof CacheStrategy_1.CacheStrategy;
};
/**
 * Root storage context.
 */
var StorageContext = /** @class */ (function () {
    function StorageContext(_a) {
        var depth = _a.depth, _b = _a.strategy, strategy = _b === void 0 ? strategy_1.LRU : _b, _c = _a.totalLeavesLimit, totalLeavesLimit = _c === void 0 ? constants_1.DEFAULT_MAX_ENTRIES_COUNT : _c, _d = _a.totalStoragesLimit, totalStoragesLimit = _d === void 0 ? Infinity : _d, _e = _a.totalLeafStoragesLimit, totalLeafStoragesLimit = _e === void 0 ? Infinity : _e, _f = _a.leavesPerStorageLimit, leavesPerStorageLimit = _f === void 0 ? Infinity : _f, _g = _a.useWeakStorage, useWeakStorage = _g === void 0 ? false : _g, _h = _a.useObjectStorage, useObjectStorage = _h === void 0 ? false : _h, _j = _a.onCreateStorage, onCreateStorage = _j === void 0 ? utils_1.noop : _j, _k = _a.onRemoveStorage, onRemoveStorage = _k === void 0 ? utils_1.noop : _k, _l = _a.onCreateLeaf, onCreateLeaf = _l === void 0 ? utils_1.noop : _l, _m = _a.onRemoveLeaf, onRemoveLeaf = _m === void 0 ? utils_1.noop : _m;
        if (!Number.isSafeInteger(depth) || depth <= 0) {
            throw new TypeError("Invalid depth, an expected natural number which is a safe integer");
        }
        var leafStrategyClass, storageStrategyClass;
        if (extendsCacheStrategy(strategy)) {
            leafStrategyClass = strategy;
            storageStrategyClass = strategy;
        }
        else {
            var config = strategy;
            leafStrategyClass = config.leafStrategyClass;
            storageStrategyClass =
                config.storageStrategyClass;
        }
        if (!extendsCacheStrategy(leafStrategyClass)) {
            throw new Error("Invalid `leafStrategyClass`");
        }
        if (!extendsCacheStrategy(storageStrategyClass)) {
            throw new Error("Invalid `storageStrategyClass`");
        }
        ensureLimitIsValid("totalLeavesLimit", totalLeavesLimit, 1);
        ensureLimitIsValid("totalStoragesLimit", totalStoragesLimit, depth);
        ensureLimitIsValid("totalLeafStoragesLimit", totalLeafStoragesLimit, 1);
        ensureLimitIsValid("leavesPerStorageLimit", leavesPerStorageLimit, 1);
        storageStrategyClass = (0, CacheStrategy_1.withDestroyable)(pickNoopIfNoLimit(storageStrategyClass, totalStoragesLimit));
        leafStrategyClass =
            totalLeavesLimit === totalLeafStoragesLimit &&
                totalLeavesLimit === leavesPerStorageLimit
                ? pickNoopIfNoLimit(leafStrategyClass, totalLeavesLimit)
                : leafStrategyClass;
        this.rootStoragesStrategy = new storageStrategyClass(totalStoragesLimit - 1);
        this.rootLeafStoragesStrategy = new RootLeafStrategy_1.RootLeafStrategy(totalLeavesLimit, new ((0, CacheStrategy_1.withDestroyable)(leafStrategyClass))(totalLeafStoragesLimit));
        this.storageClass =
            useWeakStorage || useObjectStorage ? storage_1.UnifiedStorage : storage_1.SafeMapStorage;
        this.leafStrategyClass = leafStrategyClass;
        this.leavesPerStorageLimit = leavesPerStorageLimit;
        this.params = {
            useObjectStorage: useObjectStorage,
            useWeakStorage: useWeakStorage,
            onCreateStorage: onCreateStorage,
            onRemoveStorage: onRemoveStorage,
            onCreateLeaf: onCreateLeaf,
            onRemoveLeaf: onRemoveLeaf
        };
    }
    /**
     * Instantiates new storage with supplied parents
     * @param root
     */
    StorageContext.prototype.createStorage = function (parentPaths) {
        if (parentPaths === void 0) { parentPaths = iterables_1.EMPTY_ITERABLE; }
        return new this.storageClass(this.params, parentPaths);
    };
    /**
     * Instantiates new leaf storage with supplied parents
     * @param root
     */
    StorageContext.prototype.createLeafStorage = function (parentPaths) {
        if (parentPaths === void 0) { parentPaths = iterables_1.EMPTY_ITERABLE; }
        return new LeafStorage_1.LeafStorage(this.createStorage(), new this.leafStrategyClass(this.leavesPerStorageLimit), this.params, parentPaths);
    };
    return StorageContext;
}());
exports.StorageContext = StorageContext;
//# sourceMappingURL=StorageContext.js.map